# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Rewrite the formula with basic operators visitor."""
from functools import singledispatch

from pylogics.syntax.base import And as PLTLAnd
from pylogics.syntax.base import Equivalence as PLTLEquivalence
from pylogics.syntax.base import Formula
from pylogics.syntax.base import Implies as PLTLImplies
from pylogics.syntax.base import Not as PLTLNot
from pylogics.syntax.base import Or as PLTLOr
from pylogics.syntax.base import _UnaryOp
from pylogics.syntax.pltl import Atomic as PLTLAtomic
from pylogics.syntax.pltl import (
    Before,
    Historically,
    Once,
    PropositionalFalse,
    PropositionalTrue,
    Since,
)


def rewrite_unaryop(formula: _UnaryOp):
    """Rewrite the formula for a unary operator."""
    return rewrite(formula.argument)


@singledispatch
def rewrite(formula: object) -> Formula:
    """Rewrite a formula."""
    raise NotImplementedError(
        f"handler not implemented for object of type {type(formula)}"
    )


@rewrite.register
def rewrite_true(formula: PropositionalTrue) -> Formula:
    """Compute the basic formula for the true formula."""
    return formula


@rewrite.register
def rewrite_false(formula: PropositionalFalse) -> Formula:
    """Compute the basic formula for the false formula."""
    return formula


@rewrite.register
def rewrite_atomic(formula: PLTLAtomic) -> Formula:
    """Compute the basic formula for an atomic formula."""
    return formula


@rewrite.register
def rewrite_and(formula: PLTLAnd) -> Formula:
    """Compute the basic formula for an And formula."""
    sub = [rewrite(f) for f in formula.operands]
    return PLTLAnd(*sub)


@rewrite.register
def rewrite_or(formula: PLTLOr) -> Formula:
    """Compute the basic formula for an Or formula."""
    sub = [rewrite(f) for f in formula.operands]
    return PLTLOr(*sub)


@rewrite.register
def rewrite_not(formula: PLTLNot) -> Formula:
    """Compute the basic formula for a Not formula."""
    sub = rewrite_unaryop(formula)
    return PLTLNot(sub)


@rewrite.register
def rewrite_implies(formula: PLTLImplies) -> Formula:
    """Compute the basic formula for an Implies formula."""
    head = [PLTLNot(rewrite(f)) for f in formula.operands[:-1]]
    tail = formula.operands[-1]
    return PLTLOr(*head, tail)


@rewrite.register
def rewrite_equivalence(formula: PLTLEquivalence) -> Formula:
    """Compute the basic formula for an Equivalence formula."""
    positive = PLTLAnd(*[rewrite(f) for f in formula.operands])
    negative = PLTLAnd(*[PLTLNot(rewrite(f)) for f in formula.operands])
    return PLTLOr(positive, negative)


@rewrite.register
def rewrite_before(formula: Before) -> Formula:
    """Compute the basic formula for a Before (Yesterday) formula."""
    sub = rewrite_unaryop(formula)
    return Before(sub)


@rewrite.register
def rewrite_since(formula: Since) -> Formula:
    """Compute the basic formula for a Since formula."""
    if len(formula.operands) != 2:
        head = formula.operands[0]
        tail = Since(*formula.operands[1:])
        return rewrite(Since(head, tail))
    sub = [rewrite(f) for f in formula.operands]
    return Since(*sub)


@rewrite.register
def rewrite_once(formula: Once) -> Formula:
    """Compute the basic formula for a Once formula."""
    sub = rewrite_unaryop(formula)
    return Once(sub)


@rewrite.register
def rewrite_historically(formula: Historically) -> Formula:
    """Compute the basic formula for a Historically formula."""
    return PLTLNot(Once(PLTLNot(rewrite_unaryop(formula))))



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Derived Predicates visitor."""
from functools import singledispatch
from typing import Dict, Set

from pddl.logic.base import And, Not, Or
from pddl.logic.predicates import DerivedPredicate, Predicate
from pylogics.syntax.base import And as PLTLAnd
from pylogics.syntax.base import Not as PLTLNot
from pylogics.syntax.base import Or as PLTLOr
from pylogics.syntax.pltl import Atomic as PLTLAtomic
from pylogics.syntax.pltl import (
    Before,
    Once,
    PropositionalFalse,
    PropositionalTrue,
    Since,
)
from pylogics.utils.to_string import to_string

from plan4past.helpers.utils import add_val_prefix, replace_symbols


@singledispatch
def derived_predicates(
    formula: object, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a formula."""
    raise NotImplementedError(f"handler not implemented for object {type(formula)}")


@derived_predicates.register
def derived_predicates_true(
    _formula: PropositionalTrue, _atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a true formula."""
    val = Predicate(add_val_prefix("true"))
    return {DerivedPredicate(val, And())}


@derived_predicates.register
def derived_predicates_false(
    _formula: PropositionalFalse, _atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a false formula."""
    val = Predicate(add_val_prefix("false"))
    return {DerivedPredicate(val, Or())}


@derived_predicates.register
def derived_predicates_atomic(
    formula: PLTLAtomic, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for an atomic formula."""
    val = Predicate(add_val_prefix(formula.name))
    condition = atoms_to_fluents[formula]
    return {DerivedPredicate(val, condition)}


@derived_predicates.register
def derived_predicates_and(
    formula: PLTLAnd, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a PPLTL And formula."""
    formula_name = to_string(formula)
    val = Predicate(add_val_prefix(replace_symbols(formula_name)))
    val_ops = [
        Predicate(add_val_prefix(replace_symbols(to_string(op))))
        for op in formula.operands
    ]
    condition = And(*val_ops)
    der_pred_ops = [derived_predicates(op, atoms_to_fluents) for op in formula.operands]
    return {DerivedPredicate(val, condition)}.union(*der_pred_ops)


@derived_predicates.register
def derived_predicates_or(
    formula: PLTLOr, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a PPLTL Or formula."""
    formula_name = to_string(formula)
    val = Predicate(add_val_prefix(replace_symbols(formula_name)))
    val_ops = [
        Predicate(add_val_prefix(replace_symbols(to_string(op))))
        for op in formula.operands
    ]
    condition = Or(*val_ops)
    der_pred_ops = [derived_predicates(op, atoms_to_fluents) for op in formula.operands]
    return {DerivedPredicate(val, condition)}.union(*der_pred_ops)


@derived_predicates.register
def derived_predicates_not(
    formula: PLTLNot, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a PPLTL Not formula."""
    formula_name = to_string(formula)
    val = Predicate(add_val_prefix(replace_symbols(formula_name)))
    condition = Not(
        Predicate(add_val_prefix(replace_symbols(to_string(formula.argument))))
    )
    der_pred_arg = derived_predicates(formula.argument, atoms_to_fluents)
    return {DerivedPredicate(val, condition)}.union(der_pred_arg)


@derived_predicates.register
def derived_predicates_before(
    formula: Before, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a Before formula."""
    formula_name = to_string(formula)
    val = Predicate(add_val_prefix(replace_symbols(formula_name)))
    condition = Predicate(replace_symbols(to_string(formula)))
    der_pred_arg = derived_predicates(formula.argument, atoms_to_fluents)
    return {DerivedPredicate(val, condition)}.union(der_pred_arg)


@derived_predicates.register
def derived_predicates_since(
    formula: Since, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a Since formula."""
    if len(formula.operands) != 2:
        head = formula.operands[0]
        tail = Since(*formula.operands[1:])
        return derived_predicates(Since(head, tail), atoms_to_fluents)
    formula_name = to_string(formula)
    val = Predicate(add_val_prefix(replace_symbols(formula_name)))
    op_or_1 = Predicate(add_val_prefix(replace_symbols(to_string(formula.operands[1]))))
    op_or_2 = And(
        Predicate(add_val_prefix(replace_symbols(to_string(formula.operands[0])))),
        Predicate(f"Y-{replace_symbols(to_string(formula))}"),
    )
    condition = Or(op_or_1, op_or_2)
    der_pred_ops = [derived_predicates(op, atoms_to_fluents) for op in formula.operands]
    return {DerivedPredicate(val, condition)}.union(*der_pred_ops)


@derived_predicates.register
def derived_predicates_once(
    formula: Once, atoms_to_fluents: Dict[PLTLAtomic, Predicate]
) -> Set[DerivedPredicate]:
    """Compute the derived predicate for a Once formula."""
    formula_name = to_string(formula)
    val = Predicate(add_val_prefix(replace_symbols(formula_name)))
    condition = Or(
        Predicate(add_val_prefix(replace_symbols(to_string(formula.argument)))),
        Predicate(f"Y-{replace_symbols(to_string(formula))}"),
    )
    der_pred_arg = derived_predicates(formula.argument, atoms_to_fluents)
    return {DerivedPredicate(val, condition)}.union(der_pred_arg)



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Miscellanea utilities."""
from typing import Dict

from pddl.logic import Predicate, constants
from pylogics.syntax.base import Formula
from pylogics.syntax.pltl import Atomic as PLTLAtomic

from plan4past.utils.atoms_visitor import find_atoms


def add_val_prefix(name: str):
    """Add the 'prime' prefix."""
    return "val-" + name.replace('"', "")


def remove_before_prefix(name: str):
    """Remove the 'Y' prefix."""
    return (
        name.replace("Y-", "")
        if name.startswith("Y-")
        else name.replace("Y", "", 1)
        if name.startswith("Y")
        else name
    )


def remove_val_prefix(name: str):
    """Remove the 'prime' prefix."""
    return name.replace("val-", "") if name.startswith("val-") else name


def replace_symbols(name: str):
    """Stringify symbols."""
    return (
        name.replace('"', "")
        .replace("(", "")
        .replace(")", "")
        .replace("&", "and")
        .replace("|", "or")
        .replace("~", "not-")
        .replace("!", "not-")
        .replace(" ", "-")
    )


def default_mapping(formula: Formula) -> Dict[PLTLAtomic, Predicate]:
    """Compute mapping from atoms to fluents using underscores."""
    symbols = find_atoms(formula)
    from_atoms_to_fluents = {}
    for symbol in symbols:
        name, *cons = symbol.name.split("_")
        if cons:
            from_atoms_to_fluents[symbol] = Predicate(name, *constants(" ".join(cons)))
        else:
            from_atoms_to_fluents[symbol] = Predicate(name)
    return from_atoms_to_fluents


def check_(condition: bool, message: str = "") -> None:
    """
    User-defined assert.

    This function is useful to avoid the use of the built-in assert statement, which is removed
        when the code is compiled in optimized mode. For more information, see
        https://bandit.readthedocs.io/en/1.7.5/plugins/b101_assert_used.html
    """
    if not condition:
        raise AssertionError(message)



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Predicates visitor."""
import functools
from functools import singledispatch
from typing import Set

from pddl.logic.predicates import Predicate
from pylogics.syntax.base import And as PLTLAnd
from pylogics.syntax.base import Not as PLTLNot
from pylogics.syntax.base import Or as PLTLOr
from pylogics.syntax.base import _BinaryOp, _UnaryOp
from pylogics.syntax.pltl import Atomic as PLTLAtomic
from pylogics.syntax.pltl import (
    Before,
    Once,
    PropositionalFalse,
    PropositionalTrue,
    Since,
)
from pylogics.utils.to_string import to_string

from plan4past.helpers.utils import replace_symbols


def predicates_binaryop(formula: _BinaryOp):
    """Compute predicate for a binary operator."""
    return set(functools.reduce(set.union, map(predicates, formula.operands)))  # type: ignore[arg-type]


def predicates_unaryop(formula: _UnaryOp):
    """Compute predicate for a unary operator."""
    return predicates(formula.argument)


@singledispatch
def predicates(formula: object) -> Set[Predicate]:
    """Compute predicate for a formula."""
    raise NotImplementedError(
        f"handler not implemented for object of type {type(formula)}"
    )


@predicates.register
def predicates_true(_formula: PropositionalTrue) -> Set[Predicate]:
    """Compute predicate for a true formula."""
    return set()


@predicates.register
def predicates_false(_formula: PropositionalFalse) -> Set[Predicate]:
    """Compute predicate for a false formula."""
    return set()


@predicates.register
def predicates_atomic(_formula: PLTLAtomic) -> Set[Predicate]:
    """Compute predicate for an atomic formula."""
    return set()


@predicates.register
def predicates_and(formula: PLTLAnd) -> Set[Predicate]:
    """Compute predicate for an And formula."""
    return predicates_binaryop(formula)


@predicates.register
def predicates_or(formula: PLTLOr) -> Set[Predicate]:
    """Compute predicate for an Or formula."""
    return predicates_binaryop(formula)


@predicates.register
def predicates_not(formula: PLTLNot) -> Set[Predicate]:
    """Compute predicate for a Not formula."""
    return predicates(formula.argument)


@predicates.register
def predicates_before(formula: Before) -> Set[Predicate]:
    """Compute predicate for a Before (Yesterday) formula."""
    quoted = Predicate(replace_symbols(to_string(formula)))
    sub = predicates_unaryop(formula)
    return sub.union({quoted})


@predicates.register
def predicates_since(formula: Since) -> Set[Predicate]:
    """Compute predicate for a Since formula."""
    if len(formula.operands) != 2:
        head = formula.operands[0]
        tail = Since(*formula.operands[1:])
        return predicates(Since(head, tail))
    formula_name = replace_symbols(to_string(formula))
    quoted = Predicate(f"Y-{formula_name}")
    subsinces = predicates_binaryop(formula)
    return {quoted}.union(subsinces)


@predicates.register
def predicates_once(formula: Once) -> Set[Predicate]:
    """Compute predicate for a Once formula."""
    formula_name = replace_symbols(to_string(formula))
    quoted = Predicate(f"Y-{formula_name}")
    sub = predicates_unaryop(formula)
    return sub.union({quoted})



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Find Atoms visitor."""
import functools
from functools import singledispatch
from typing import Set

from pylogics.syntax.base import And as PLTLAnd
from pylogics.syntax.base import Not as PLTLNot
from pylogics.syntax.base import Or as PLTLOr
from pylogics.syntax.base import _BinaryOp, _UnaryOp
from pylogics.syntax.pltl import Atomic as PLTLAtomic
from pylogics.syntax.pltl import Before
from pylogics.syntax.pltl import FalseFormula as PLTLFalse
from pylogics.syntax.pltl import Historically, Once, Since
from pylogics.syntax.pltl import TrueFormula as PLTLTrue


def find_atoms_binaryop(formula: _BinaryOp):
    """Find atoms for a binary operator."""
    return set(functools.reduce(set.union, map(find_atoms, formula.operands)))  # type: ignore[arg-type]


def find_atoms_unaryop(formula: _UnaryOp):
    """Find atoms for a unary operator."""
    return find_atoms(formula.argument)


@singledispatch
def find_atoms(obj: object) -> Set[PLTLAtomic]:
    """Find atoms for a formula."""
    raise ValueError(f"object of type {type(obj)} is not supported by this function")


@find_atoms.register
def find_atoms_true(_formula: PLTLTrue) -> Set[PLTLAtomic]:
    """Find atoms for a true formula."""
    return {PLTLAtomic("true")}


@find_atoms.register
def find_atoms_false(_formula: PLTLFalse) -> Set[PLTLAtomic]:
    """Find atoms for a false formula."""
    return {PLTLAtomic("false")}


@find_atoms.register
def find_atoms_atom(formula: PLTLAtomic) -> Set[PLTLAtomic]:
    """Find atoms for an atomic formula."""
    return {formula}


@find_atoms.register
def find_atoms_and(formula: PLTLAnd) -> Set[PLTLAtomic]:
    """Find atoms for an And formula."""
    return find_atoms_binaryop(formula)


@find_atoms.register
def find_atoms_or(formula: PLTLOr) -> Set[PLTLAtomic]:
    """Find atoms for an Or formula."""
    return find_atoms_binaryop(formula)


@find_atoms.register
def find_atoms_not(formula: PLTLNot) -> Set[PLTLAtomic]:
    """Find atoms for a Not formula."""
    return find_atoms_unaryop(formula)


@find_atoms.register
def find_atoms_before(formula: Before) -> Set[PLTLAtomic]:
    """Find atoms for a Before formula."""
    return find_atoms_unaryop(formula)


@find_atoms.register
def find_atoms_since(formula: Since) -> Set[PLTLAtomic]:
    """Find atoms for a Since formula."""
    return find_atoms_binaryop(formula)


@find_atoms.register
def find_atoms_once(formula: Once) -> Set[PLTLAtomic]:
    """Find atoms for a Once formula."""
    return find_atoms_unaryop(formula)


@find_atoms.register
def find_atoms_historically(formula: Historically) -> Set[PLTLAtomic]:
    """Find atoms for a Historically formula."""
    return find_atoms_unaryop(formula)



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Mapping parser."""
import re
from typing import Dict, Optional, Sequence, Set, Tuple

from pddl.custom_types import name
from pddl.logic import Constant, Predicate, constants
from pylogics.syntax.base import Formula
from pylogics.syntax.pltl import Atomic as PLTLAtomic

from plan4past.utils.atoms_visitor import find_atoms


class MappingParserError(Exception):
    """Mapping parser error."""

    def __init__(self, message: str, *args, row_id: Optional[int] = None) -> None:
        """
        Initialize a mapping parser error.

        :param row_id: the row id
        """
        self.row_id = row_id

        super().__init__(self._make_message_prefix() + message, *args)

    def _make_message_prefix(self) -> str:
        """Make the message prefix."""
        if self.row_id is None:
            return "invalid mapping: "
        return f"invalid mapping at row {self.row_id}: "


SYMBOL_REGEX = re.compile("[a-z_]([a-zA-Z0-9_-]+[a-zA-Z0-9_])|[a-z_][a-zA-Z0-9_]*")
"""The following is a sub-regex of AtomName.REGEX in pylogics.syntax.base."""


def _check_pddl_name(row_id: int, pddl_name: str, what: str) -> None:
    """Check that a PDDL name is valid."""
    try:
        name(pddl_name)
    except ValueError as e:
        raise MappingParserError(
            f"got an invalid name for {what}: '{pddl_name}'. It must match the regex {name.REGEX}",
            row_id=row_id,
        ) from e


def _parse_constants(row_id: int, constants_str: str) -> Sequence[Constant]:
    """
    Parse a string of constants.

    :param row_id: the row id
    :param constants_str: the string of constants
    :return: the sequence of constants
    """
    try:
        return constants(constants_str)
    except ValueError as e:
        raise MappingParserError(f"got an invalid constant: {e}", row_id=row_id) from e


def _parse_predicate(row_id: int, predicate_str: str) -> Predicate:
    """
    Parse a predicate.

    :param predicate_str: the predicate string
    :return: the predicate
    """
    predicate_name, cons = predicate_str.split(" ", maxsplit=1)

    _check_pddl_name(row_id, predicate_name, "a predicate")
    parsed_constants = _parse_constants(row_id, cons)
    return Predicate(predicate_name, *parsed_constants)


def _process_row(row_id: int, row_str: str) -> Tuple[str, Predicate]:
    """
    Process a row of the mapping file.

    :param row_id: the row id
    :param row_str: the row string
    :return: the pair (symbol_name, predicate_str)
    """
    # check the row is a valid mapping
    comma_separated_row_parts = row_str.split(",")
    if len(comma_separated_row_parts) != 2:
        raise MappingParserError(
            "expected a mapping of the form '<symbol>,<predicate>'", row_id=row_id
        )

    symbol_name, predicate = comma_separated_row_parts

    # strip leading and trailing whitespaces
    symbol_name = symbol_name.strip()
    predicate_str = predicate.strip()

    if symbol_name == "" or SYMBOL_REGEX.fullmatch(symbol_name) is None:
        raise MappingParserError(
            "symbol cannot be empty string and must match the regex {SYMBOL_REGEX}",
            row_id=row_id,
        )
    if predicate_str == "":
        raise MappingParserError("predicate cannot be empty string", row_id=row_id)

    predicate = _parse_predicate(row_id, predicate_str)
    return symbol_name, predicate


def _check_unmapped_symbols(
    all_symbol_names: Set[str], mapped_symbol_names: Set[str]
) -> None:
    """Check that all symbols are mapped."""
    # check if there are unmapped symbols (exclude 'true' and 'false')
    unmapped_symbols = all_symbol_names - {"true", "false"} - mapped_symbol_names
    # if some symbols are not mapped, raise an error
    if unmapped_symbols:
        raise MappingParserError(
            f"the following symbols of the formula are not mapped: {unmapped_symbols}"
        )


def should_skip_row(row_str: str) -> bool:
    """Check if a row should be skipped."""
    # skip empty lines
    if row_str.strip() == "":
        return True

    # skip comments
    if row_str.strip().startswith(";"):
        return True

    return False


def mapping_parser(text: str, formula: Formula) -> Dict[PLTLAtomic, Predicate]:
    """Parse symbols to ground predicates mapping."""
    symbols = find_atoms(formula)
    symbols_by_name = {symbol.name: symbol for symbol in symbols}
    maps = text.splitlines(keepends=False)
    from_atoms_to_fluents = {}
    mapped_symbol_names = set()

    for row_id, vmap in enumerate(maps):
        if should_skip_row(vmap):
            continue

        symbol_name, predicate = _process_row(row_id, vmap)

        if symbol_name not in symbols_by_name:
            # don't need to process this row, since the symbol does not occur in the formula
            continue

        if symbol_name in mapped_symbol_names:
            raise MappingParserError(
                f"symbol '{symbol_name}' is mapped multiple times", row_id=row_id
            )

        symbol = symbols_by_name[symbol_name]
        mapped_symbol_names.add(symbol_name)

        from_atoms_to_fluents[symbol] = predicate

    _check_unmapped_symbols(set(symbols_by_name.keys()), mapped_symbol_names)
    return from_atoms_to_fluents



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Value predicates for val predicates visitor."""
import functools
from functools import singledispatch
from typing import Set

from pddl.logic.predicates import Predicate
from pylogics.syntax.base import And as PLTLAnd
from pylogics.syntax.base import Not as PLTLNot
from pylogics.syntax.base import Or as PLTLOr
from pylogics.syntax.base import _BinaryOp, _UnaryOp
from pylogics.syntax.pltl import Atomic as PLTLAtomic
from pylogics.syntax.pltl import (
    Before,
    Once,
    PropositionalFalse,
    PropositionalTrue,
    Since,
)
from pylogics.utils.to_string import to_string

from plan4past.helpers.utils import add_val_prefix, replace_symbols


def val_predicates_binaryop(formula: _BinaryOp):
    """Compute the value predicate for a binary operator formula."""
    return set(functools.reduce(set.union, map(val_predicates, formula.operands)))  # type: ignore[arg-type]


def val_predicates_unaryop(formula: _UnaryOp):
    """Compute the value predicate for a unary operator formula."""
    return val_predicates(formula.argument)


@singledispatch
def val_predicates(formula: object) -> Set[Predicate]:
    """Compute the value predicate for a formula."""
    raise NotImplementedError(
        f"handler not implemented for object of type {type(formula)}"
    )


@val_predicates.register
def val_predicates_true(_formula: PropositionalTrue) -> Set[Predicate]:
    """Compute the value predicate for a true formula."""
    return {Predicate(add_val_prefix("true"))}


@val_predicates.register
def val_predicates_false(_formula: PropositionalFalse) -> Set[Predicate]:
    """Compute the value predicate for a false formula."""
    return {Predicate(add_val_prefix("false"))}


@val_predicates.register
def val_predicates_atomic(formula: PLTLAtomic) -> Set[Predicate]:
    """Compute the value predicate for an atomic formula."""
    return {Predicate(add_val_prefix(formula.name))}


@val_predicates.register
def val_predicates_and(formula: PLTLAnd) -> Set[Predicate]:
    """Compute the value predicate for an And formula."""
    formula_name = replace_symbols(to_string(formula))
    value = Predicate(add_val_prefix(formula_name))
    subands = val_predicates_binaryop(formula)
    return {value}.union(subands)


@val_predicates.register
def val_predicates_or(formula: PLTLOr) -> Set[Predicate]:
    """Compute the value predicate for an Or formula."""
    formula_name = replace_symbols(to_string(formula))
    value = Predicate(add_val_prefix(formula_name))
    subors = val_predicates_binaryop(formula)
    return {value}.union(subors)


@val_predicates.register
def val_predicates_not(formula: PLTLNot) -> Set[Predicate]:
    """Compute the value predicate for a Not formula."""
    formula_name = replace_symbols(to_string(formula))
    value = Predicate(add_val_prefix(formula_name))
    sub = val_predicates_unaryop(formula)
    return sub.union({value})


@val_predicates.register
def val_predicates_before(formula: Before) -> Set[Predicate]:
    """Compute the value predicate for a Before (Yesterday) formula."""
    formula_name = replace_symbols(to_string(formula))
    value = Predicate(add_val_prefix(formula_name))
    sub = val_predicates_unaryop(formula)
    return sub.union({value})


@val_predicates.register
def val_predicates_since(formula: Since) -> Set[Predicate]:
    """Compute the value predicate for a Since formula."""
    if len(formula.operands) != 2:
        head = formula.operands[0]
        tail = Since(*formula.operands[1:])
        return val_predicates(Since(head, tail))
    formula_name = replace_symbols(to_string(formula))
    value = Predicate(add_val_prefix(formula_name))
    subsinces = val_predicates_binaryop(formula)
    return {value}.union(subsinces)


@val_predicates.register
def val_predicates_once(formula: Once) -> Set[Predicate]:
    """Compute the value predicate for a Once formula."""
    formula_name = replace_symbols(to_string(formula))
    value = Predicate(add_val_prefix(formula_name))
    sub = val_predicates_unaryop(formula)
    return sub.union({value})



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#

"""Compiler from PDDL Domain and PPLTL into a new PDDL domain."""
from typing import AbstractSet, Dict, Optional, Set, Tuple

from pddl.core import Action, Domain, Problem, Requirements
from pddl.logic import Constant
from pddl.logic.base import And, Not
from pddl.logic.effects import AndEffect, When
from pddl.logic.predicates import DerivedPredicate, Predicate
from pylogics.syntax.base import Formula, Logic
from pylogics.syntax.pltl import Atomic as PLTLAtomic
from pylogics.utils.to_string import to_string

from plan4past.helpers.utils import (
    add_val_prefix,
    check_,
    default_mapping,
    remove_before_prefix,
    replace_symbols,
)
from plan4past.utils.atoms_visitor import find_atoms
from plan4past.utils.derived_visitor import derived_predicates
from plan4past.utils.predicates_visitor import predicates
from plan4past.utils.rewrite_formula_visitor import rewrite
from plan4past.utils.val_predicates_visitor import val_predicates


class Compiler:
    """Compiler of PPLTL goals into PDDL."""

    def __init__(
        self,
        domain: Domain,
        problem: Problem,
        formula: Formula,
        from_atoms_to_fluent: Optional[Dict[PLTLAtomic, Predicate]] = None,
    ) -> None:
        """
        Initialize the compiler.

        :param domain: the domain
        :param problem: the problem
        :param formula: the formula
        :param from_atoms_to_fluent: optional mapping from atoms to fluent
        """
        self.domain = domain
        self.problem = problem
        self.formula = rewrite(formula)
        if from_atoms_to_fluent:
            self.from_atoms_to_fluent = from_atoms_to_fluent
            self.validate_mapping(domain, formula, from_atoms_to_fluent)
        else:
            self.from_atoms_to_fluent = default_mapping(self.formula)

        check_(self.formula.logic == Logic.PLTL, "only PPLTL is supported!")

        self._executed: bool = False
        self._result_domain: Optional[Domain] = None
        self._result_problem: Optional[Problem] = None

        self._derived_predicates: Set[DerivedPredicate] = set()

    @classmethod
    def validate_mapping(
        cls,
        _domain: Domain,
        _formula: Formula,
        from_atoms_to_fluent: Dict[PLTLAtomic, Predicate],
    ):
        """
        Check that the mapping is valid wrt the problem instance.

        In particular:
        - check that all the formula atoms are covered (TODO)
        - check that all the atoms are legal wrt the formula
        - check that the fluents are legal wrt the domain

        :param _domain:
        :param _formula:
        :param from_atoms_to_fluent:
        :return:
        """
        for _atom, fluent in from_atoms_to_fluent.items():
            check_(all(isinstance(t, Constant) for t in fluent.terms))

    @property
    def result(self) -> Tuple[Domain, Problem]:
        """Get the result."""
        if self._result_domain and self._result_problem is None:
            raise ValueError("compilation not executed yet")
        return self._result_domain, self._result_problem

    def compile(self):
        """Compute the new domain and the new problem."""
        if not self._executed:
            self._compile_domain()
            self._compile_problem()
            self._executed = True

    def _compile_domain(self):
        """Compute the new domain."""
        new_predicates = predicates(self.formula).union(val_predicates(self.formula))
        new_derived_predicates = derived_predicates(
            self.formula, self.from_atoms_to_fluent
        )
        new_whens = _compute_whens(self.formula)
        domain_actions = _update_domain_actions_det(self.domain.actions, new_whens)

        self._result_domain = Domain(
            name=self.domain.name,
            requirements=[
                *self.domain.requirements,
                Requirements.DERIVED_PREDICATES,
                Requirements.CONDITIONAL_EFFECTS,
                Requirements.NEG_PRECONDITION,
            ],
            types=self.domain.types,
            constants=self.domain.constants,
            predicates=[*self.domain.predicates, *new_predicates],
            derived_predicates=[
                *self.domain.derived_predicates,
                *new_derived_predicates,
            ],
            actions=domain_actions,
        )

    def _compile_problem(self):
        """Compute the new problem."""
        new_init = (
            {*self.problem.init, Predicate("true")}
            if PLTLAtomic("true") in find_atoms(self.formula)
            else set(self.problem.init)
        )

        self._result_problem = Problem(
            name=self.problem.name,
            domain_name=self.domain.name,
            requirements=self.problem.requirements,
            objects=[*self.problem.objects],
            init=new_init,
            goal=And(
                Predicate(add_val_prefix(replace_symbols(to_string(self.formula))))
            ),
        )


def _compute_whens(formula: Formula) -> Set[When]:
    """Compute conditional effects for formula progression."""
    return {
        When(Predicate(add_val_prefix(remove_before_prefix(p.name))), p)
        for p in predicates(formula)
    }.union(
        When(Not(Predicate(add_val_prefix(remove_before_prefix(p.name)))), Not(p))
        for p in predicates(formula)
    )


def _update_domain_actions_det(
    actions: AbstractSet[Action], progression: Set[When]
) -> Set[Action]:
    """Update domain actions."""
    new_actions = set()
    for action in actions:
        if isinstance(action.effect, AndEffect):
            previous_effects = action.effect.operands
            new_actions.add(
                Action(
                    name=action.name,
                    parameters=[*action.parameters],
                    precondition=And(action.precondition),
                    effect=AndEffect(*previous_effects, *progression),
                )
            )
        else:
            new_actions.add(
                Action(
                    name=action.name,
                    parameters=[*action.parameters],
                    precondition=And(action.precondition),
                    effect=AndEffect(action.effect, *progression),
                )
            )
    return new_actions



# -*- coding: utf-8 -*-
#
# Copyright 2021 -- 2023 WhiteMech
#
# ------------------------------
#
# This file is part of Plan4Past.
#
# Plan4Past is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Plan4Past is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Plan4Past.  If not, see <https://www.gnu.org/licenses/>.
#


"""Main entrypoint for Plan4Past."""
from pathlib import Path
from typing import Tuple

import click
from pddl.core import Domain, Problem
from pddl.formatter import domain_to_string, problem_to_string
from pddl.parser.domain import DomainParser
from pddl.parser.problem import ProblemParser
from pylogics.parsers import parse_pltl
from pylogics.syntax.base import Formula

from plan4past.compiler import Compiler
from plan4past.utils.mapping_parser import mapping_parser

DEFAULT_NEW_DOMAIN_FILENAME: str = "new-domain.pddl"
DEFAULT_NEW_PROBLEM_FILENAME: str = "new-problem.pddl"


@click.command()
@click.option(
    "-d",
    "--domain",
    required=True,
    help="Path to PDDL domain file.",
    type=click.Path(exists=True, readable=True),
)
@click.option(
    "-p",
    "--problem",
    required=True,
    help="Path to PDDL problem file.",
    type=click.Path(exists=True, readable=True),
)
@click.option("-g", "--goal-inline", help="The PPLTL goal formula.", type=str)
@click.option(
    "-gf",
    "--goal-file",
    help="The path to the PPLTL goal formula.",
    type=click.Path(exists=True, readable=True),
)
@click.option(
    "-m",
    "--mapping",
    help="The mapping file.",
    type=click.Path(exists=True, readable=True),
    default=None,
)
@click.option(
    "-od",
    "--out-domain",
    default=f"./{DEFAULT_NEW_DOMAIN_FILENAME}",
    help="Path to PDDL file to store the new domain.",
    type=click.Path(dir_okay=False),
)
@click.option(
    "-op",
    "--out-problem",
    default=f"./{DEFAULT_NEW_PROBLEM_FILENAME}",
    help="Path to PDDL file to store the new problem.",
    type=click.Path(dir_okay=False),
)
def cli(domain, problem, goal_inline, goal_file, mapping, out_domain, out_problem):
    """Plan4Past: Planning for Pure-Past Temporally Extended Goals."""
    goal = _get_goal(goal_inline, goal_file)

    in_domain, in_problem, formula = _parse_instance(domain, problem, goal)

    var_map = (
        mapping_parser(Path(mapping).read_text(encoding="utf-8"), formula)
        if mapping
        else None
    )

    compiled_domain, compiled_problem = _compile_instance(
        in_domain, in_problem, formula, var_map
    )

    try:
        with open(out_domain, "w+", encoding="utf-8") as d:
            d.write(domain_to_string(compiled_domain))
        with open(out_problem, "w+", encoding="utf-8") as p:
            p.write(problem_to_string(compiled_problem))
    except Exception as e:
        raise IOError(
            "[ERROR]: Something wrong occurred while writing the compiled domain and problem."
        ) from e


def _get_goal(goal_inline, goal_file) -> str:
    """Get the goal formula."""
    if goal_inline and goal_file:
        raise ValueError(
            "[ERROR]: You cannot specify both the goal formula and the goal file."
        )
    if goal_file:
        return Path(goal_file).read_text(encoding="utf-8")
    if goal_inline:
        return goal_inline
    raise ValueError(
        "[ERROR]: You must specify either the goal formula or the goal file."
    )


def _parse_instance(in_domain, in_problem, goal) -> Tuple[Domain, Problem, Formula]:
    """Parse the PDDL domain and problem files and the PPLTL goal formula."""
    domain_parser = DomainParser()
    problem_parser = ProblemParser()

    domain = domain_parser(Path(in_domain).read_text(encoding="utf-8"))
    problem = problem_parser(Path(in_problem).read_text(encoding="utf-8"))
    formula = parse_pltl(goal)

    return domain, problem, formula


def _compile_instance(domain, problem, formula, mapping) -> Tuple[Domain, Problem]:
    """Compile the PDDL domain and problem files and the PPLTL goal formula."""
    compiler = Compiler(domain, problem, formula, mapping)
    compiler.compile()
    compiled_domain, compiled_problem = compiler.result

    return compiled_domain, compiled_problem


if __name__ == "__main__":
    cli()  # pragma: no cover
